<?php

namespace Groundhogg;

use function Groundhogg\Notices\add_notice;
use function Groundhogg\Notices\wp_redirect_with_notice;

if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

include __DIR__ . '/managed-page.php';

if ( ! function_exists( 'obfuscate_email' ) ):
	/**
	 * Obfuscate an email address
	 *
	 * @param $email
	 *
	 * @return string|string[]|null
	 */
	function obfuscate_email( $email ) {
		if ( ! is_email( $email ) ) {
			return false;
		}

		$parts = explode( '@', $email );

		$parts[0] = preg_replace( '/(?!^).(?=.+$)/', '*', $parts[0] );
		$parts[1] = preg_replace( '/(?!^).(?=.+\.)/', '*', $parts[1] );
		$email    = implode( '@', $parts );

		return $email;
	}

endif;

if ( ! function_exists( __NAMESPACE__ . '\mail_gdpr_data' ) ) {

	/**
	 * Mail the contact profile to the contact which requested it.
	 * Uses the regular wp_mail function.
	 *
	 * @param $contact_id
	 *
	 * @return bool
	 */
	function mail_gdpr_data( $contact_id ) {

		$contact = get_contactdata( $contact_id );

		ob_start();

		do_action( 'groundhogg/preferences/gdpr_audit_message/before', $contact );

		?>
        <p><?php _e( 'You are receiving this message because you have requested an audit of your personal information. This message contains all current information about your contact profile.', 'groundhogg' ) ?></p>
        <h3><?php _e( 'Basic Information', 'groundhogg' ) ?></h3>
		<?php

		html()->list_table( [
			'style' => [
				'border-spacing' => '10px'
			]
		], [], [
			[ __( 'Name' ), $contact->get_full_name() ],
			[ __( 'Email' ), $contact->get_email() ],
			[ __( 'Phone' ), $contact->get_phone_number() ],
			[ __( 'Mobile' ), $contact->get_mobile_number() ],
			[ __( 'Address' ), implode( ', ', $contact->get_address() ) ],
			[ __( 'IP Address' ), $contact->get_ip_address() ],
			[ __( 'Subscribed' ), date_i18n( get_date_time_format(), date_as_int( $contact->get_date_created() ) ) ],
			[ __( 'Tags' ), implode( ', ', $contact->get_tags( true ) ) ],
		] );

		do_action( 'groundhogg/preferences/gdpr_audit_message/after_basic', $contact );

		?>
        <h3><?php _e( 'Other Information', 'groundhogg' ) ?></h3>
		<?php

		$properties = Properties::instance()->get_fields();

		html()->list_table( [
			'style' => [
				'border-spacing' => '10px'
			]
		], [], array_map( function ( $p ) use ( $contact ) {

			return [
				$p['label'],
				display_custom_field( $p['name'], $contact, false )
			];

		}, $properties ) );

		do_action( 'groundhogg/preferences/gdpr_audit_message/after_other', $contact );

		$files = $contact->get_files();

		if ( ! empty( $files ) ):

			?>
            <h3><?php _e( 'Files', 'groundhogg' ) ?></h3>
            <ul>
			<?php

			foreach ( $files as $i => $file ) {
				printf( '<li><a href="%s">%s</a></li>', esc_url( permissions_key_url( $file['url'], $contact, 'download_files' ) . '&identity=' . encrypt( $contact->get_email() ) ), esc_html( $file['name'] ) );
			}

			?></ul><?php

		endif;

		$contact_methods = [
			html()->e( 'a', [ 'href' => 'mailto: ' . get_default_from_email() ], get_default_from_email() ),
		];

		$phone = get_option( 'gh_business_phone' );

		if ( $phone ) {
			$contact_methods[] = html()->e( 'a', [ 'href' => 'tel: ' . $phone ], $phone );
		}

		?>
        <p style="margin-top: 30px"><?php echo get_option( 'gh_business_name' ) ?></p>
        <p><?php echo implode( ' | ', $contact_methods ) ?></p>
        <p>
            <i><?php _e( 'This information is provided without any guarantee of being accurate, and is exhaustive to the best of our knowledge, with the exception of potentially sensitive information.', 'groundhogg' ) ?></i>
        </p>
        <p>
            <i><?php _e( 'If there is information that you wish to obtain that is not included in this audit, please contact us.', 'groundhogg' ) ?></i>
        </p>
		<?php

		$message = ob_get_clean();

		// Filters the message
		$message = apply_filters( 'groundhogg/preferences/gdpr_audit_message', $message );

		$subject_line = sprintf( __( '[%s] Your personal profile audit', 'groundhogg' ), get_option( 'gh_business_name' ) );

		// Filters the GDPR audit subject line
		$subject_line = apply_filters( 'groundhogg/preferences/gdpr_audit_subject_line', $subject_line );

		return \Groundhogg_Email_Services::send_transactional( $contact->get_email(), wp_specialchars_decode( $subject_line ), $message, [
			'Content-Type: text/html'
		] );
	}
}

if ( ! function_exists( __NAMESPACE__ . '\send_email_preferences_link' ) ) {

	/**
	 * Send an email notification to the contact with a link that will allow them to access the preferences center.
	 *
	 * @return bool
	 */
	function send_email_preferences_link() {

		$email = get_post_var( 'email' );
		$email = sanitize_email( $email );

		if ( ! $email ) {
			return false;
		}

		$contact = get_contactdata( $email );

		if ( ! is_a_contact( $contact ) ) {
			return false;
		}

		$preferences_link = managed_page_url( 'preferences/manage' );
		$preferences_link = permissions_key_url( $preferences_link, $contact, 'preferences' );
		$preferences_link = add_query_arg( 'identity', encrypt( $email ), $preferences_link );

		$message = __( 'Someone has requested to manage your email preferences:' ) . "\r\n\r\n";
		/* translators: %s: Site name. */
		$message .= sprintf( __( 'Site Name: %s', 'groundhogg' ), get_bloginfo( 'name' ) ) . "\r\n\r\n";
		$message .= __( 'If this was a mistake, just ignore this email and nothing will happen.', 'groundhogg' ) . "\r\n\r\n";
		$message .= __( 'To manage your preferences, visit the following address:', 'groundhogg' ) . "\r\n\r\n";
		$message .= $preferences_link;

		$subject = sprintf( _x( '[%s] Manage your preferences', 'subject line', 'groundhogg' ), get_bloginfo( 'name' ) );

		/**
		 * Filters the subject
		 */
		$subject = apply_filters( 'groundhogg/preferences/send_preferences_link_subject_line', $subject, $contact );

		/**
		 * Filters the message
		 */
		$message = apply_filters( 'groundhogg/preferences/send_preferences_link_message', $message, $contact );

		add_action( 'phpmailer_init', function ( $mailer ) use ( $message ) {
			// set AltBody
			$mailer->AltBody = $message;
		} );

		return \Groundhogg_Email_Services::send_transactional(
			$contact->get_email(),
			wp_specialchars_decode( $subject ),
			make_clickable( wpautop( $message ) ),
			[
				'Content-Type: text/html'
			] );
	}
}

$contact         = get_contactdata();
$permissions_key = get_permissions_key();

if ( $permissions_key && ( $enc_identity = get_url_var( 'identity' ) ) ) {
	$identity = decrypt( $enc_identity );
	$contact  = get_contactdata( $identity );

	// If the identity passed is valid and we can validate the permissions key we're good!
	if ( is_a_contact( $contact ) && check_permissions_key( $permissions_key, $contact, 'preferences' ) ) {
		tracking()->start_tracking( $contact );
	}
}

$action = get_query_var( 'action', 'profile' );

if ( ! is_ignore_user_tracking_precedence_enabled() ) {
	// If the user takes precedence of the tracking cookie
	// => The current user is logged in
	// => the pk is valid
	$can_edit_preferences = is_user_logged_in() || check_permissions_key( $permissions_key, $contact, 'preferences' );
} else {
	// if the contact takes precedence over the tracking cookie
	// => current user and contact match
	// => it's a site admin and they can edit contacts
	// => the pk is valid
	$can_edit_preferences = current_user_can( 'edit_contacts' ) || current_contact_and_logged_in_user_match() || check_permissions_key( $permissions_key, $contact, 'preferences' );
}

$can_edit_preferences = apply_filters( 'groundhogg/can_edit_preferences', $can_edit_preferences );

// if the visitor can't change the preferences show a default message.
if ( ! $can_edit_preferences || ! is_a_contact( $contact ) ) {
	$action = 'unidentified';
}

if ( current_user_can( 'view_contacts' ) && ! tracking()->tracking_cookie_matches_logged_in_user() ) {
	add_notice( 'notice_admin_logged_in_testing_warning' );
}

switch ( $action ):
	case 'unidentified':

		if ( wp_verify_nonce( get_post_var( '_wpnonce' ), 'request_email_confirmation' ) ) {

			if ( send_email_preferences_link() ) {
				add_notice( 'notice_preferences_link_sent' );
			} else {
				add_notice( 'notice_general_issue_message' );
			}

		}

		managed_page_head( __( 'Manage Preferences', 'groundhogg' ), 'manage' );
		?>
        <form method="post">
			<?php wp_nonce_field( 'request_email_confirmation' ); ?>
            <h3><?php _e( 'Whoops! We were unable to confirm your identity.', 'groundhogg' ); ?></h3>
            <p><?php _e( 'This may have occurred if you clicked an expired link or your browser is blocking cookies.', 'groundhogg' ); ?></p>
            <p><?php _e( 'If you are trying to change your email preferences please enter your email address and we will send you an email with a special link.', 'groundhogg' ); ?></p>
            <p>
                <label><?php _e( 'Your Email Address', 'groundhogg' ); ?>
                    <input type="email" name="email" required>
                </label>
            </p>
            <p>
                <button type="submit" class="button"><?php _e( 'Submit' ) ?></button>
            </p>
        </form>
		<?php

		managed_page_footer();

		break;
	case 'download':

		if ( wp_verify_nonce( get_request_var( '_wpnonce' ), 'download_profile' ) ) {
			if ( mail_gdpr_data( $contact->get_id() ) ) {

				$notice = 'notice_gdpr_email_sent';

				/**
				 * After the request is made to download the profile
				 *
				 * @param $contact Contact
				 */
				do_action( 'groundhogg/preferences/download_profile', $contact );

			} else {
				$notice = 'notice_general_issue_message';
			}
		}

		wp_redirect_with_notice( managed_page_url( 'preferences/profile/' ), $notice );
		die();

	default:
	case 'profile':

		if ( wp_verify_nonce( get_request_var( '_wpnonce' ), 'update_contact_profile' ) ) {

			$email = sanitize_email( get_request_var( 'email' ) );
			if ( ! $email || $email !== $contact->get_email() ) {
				add_notice( 'notice_email_verification_required' );
			} else {
				$args = [
					'first_name' => sanitize_text_field( get_request_var( 'first_name' ) ),
					'last_name'  => sanitize_text_field( get_request_var( 'last_name' ) ),
				];

				$args = apply_filters( 'groundhogg/preferences/update_profile', $args, $contact );

				if ( $contact->update( $args ) ) {
					add_notice( 'notice_profile_updated' );
				}

				apply_filters( 'groundhogg/preferences/update_profile', $args, $contact );
			}
		}

		if ( wp_verify_nonce( get_request_var( '_wpnonce' ), 'update_gdpr_consent' ) ) {

			$processing_consent = get_post_var( 'gdpr_consent' );
			$marketing_consent  = get_post_var( 'marketing_consent' );

			if ( $processing_consent !== 'yes' ) {
				$contact->revoke_gdpr_consent();
			} else {
				$contact->set_gdpr_consent();
			}

			if ( $marketing_consent !== 'yes' ) {
				$contact->revoke_gdpr_consent( 'marketing' );
			} else {
				$contact->set_gdpr_consent( 'marketing' );
			}

		}

		managed_page_head( __( 'Update Profile', 'groundhogg' ), 'profile' );

		?>
        <form action="" id="preferences" method="post">
            <p>
                <b><?php printf( __( 'Update information for <span class="contact-name">%s (%s)</span>.', 'groundhogg' ), $contact->get_full_name(), obfuscate_email( $contact->get_email() ) ) ?></b>
            </p>
            <p><?php _e( 'Use the form below to update your information to the most current.', 'groundhogg' ) ?></p>
			<?php wp_nonce_field( 'update_contact_profile' ); ?>
            <p>
                <label><?php _e( get_default_field_label( 'first_name' ) ); ?>
                    <input type="text" name="first_name" required>
                </label>
            </p>
            <p>
                <label><?php _e( get_default_field_label( 'last_name' ) ); ?>
                    <input type="text" name="last_name" required>
                </label>
            </p>
            <p>
                <label><?php _e( 'Confirm Email Address', 'groundhogg' ); ?>
                    <input type="email" name="email" required>
                </label>
            </p>
			<?php do_action( 'groundhogg/preferences/profile_form' ); ?>
            <p>
                <input id="submit" type="submit" class="button"
                       value="<?php esc_attr_e( 'Save Changes', 'groundhogg' ); ?>">
            </p>
        </form>
        <div class="box">
            <p><?php _e( 'Click below to manage your communication preferences and determine when and how you would like to receive communication from us.', 'groundhogg' ) ?></p>
            <p>
                <a id="gotopreferences" class="button"
                   href="<?php echo esc_url( managed_page_url( 'preferences/manage/' ) ); ?>"><?php _e( 'Change Email Preferences', 'groundhogg' ) ?></a>
            </p>
        </div>
		<?php if ( Plugin::$instance->preferences->is_gdpr_enabled() ): ?>
        <div class="box">
            <p><?php _e( 'If you wish to update your consent please do so below. Changes to your consent will be honored instantly.', 'groundhogg' ) ?></p>
            <form class="gdpr-consent" method="post">
				<?php wp_nonce_field( 'update_gdpr_consent' ); ?>
                <p>
					<?php echo html()->checkbox( [
						'label'   => get_default_field_label( 'gdpr_consent' ),
						'name'    => 'gdpr_consent',
						'id'      => 'gdpr_consent',
						'class'   => 'gh-gdpr',
						'value'   => 'yes',
						'title'   => _x( 'I Consent', 'form_default', 'groundhogg' ),
						'checked' => $contact->get_meta( 'gdpr_consent' ) === 'yes',
					] ) ?>
                </p>
                <p>
					<?php echo html()->checkbox( [
						'label'   => get_default_field_label( 'marketing_consent' ),
						'name'    => 'marketing_consent',
						'id'      => 'marketing_consent',
						'class'   => 'gh-gdpr',
						'value'   => 'yes',
						'title'   => _x( 'I Consent', 'form_default', 'groundhogg' ),
						'checked' => $contact->get_meta( 'marketing_consent' ) === 'yes',
					] ) ?>
                </p>
                <button class="button" type="submit"><?php _e( 'Update Consent', 'groundhogg' ); ?></button>
            </form>
            <p><?php _e( 'Click below to email yourself an audit of all personal information currently on file. Or if you wish for us to no longer have access to this information you can request a data erasure in accordance with your privacy rights.', 'groundhogg' ) ?></p>
            <p>
                <a id="downloadprofile" class="button"
                   href="<?php echo esc_url( wp_nonce_url( managed_page_url( 'preferences/download/' ), 'download_profile' ) ); ?>"><?php _e( 'Download Profile', 'groundhogg' ) ?></a>
                <a id="eraseprofile" class="button danger right"
                   href="<?php echo esc_url( managed_page_url( 'preferences/erase/' ) ); ?>"><?php _e( 'Erase Profile', 'groundhogg' ) ?></a>
            </p>
        </div>
	<?php endif; ?>
		<?php do_action( 'groundhogg/preferences/profile_form/after' ); ?>
		<?php

		managed_page_footer();
		break;

	case 'manage':

		do_action( 'groundhogg/preferences/manage/before' );

		if ( wp_verify_nonce( get_request_var( '_wpnonce' ), 'manage_email_preferences' ) ) {

			$preference = get_request_var( 'preference' );
			$redirect   = false;
			$notice     = 'notice_preferences_updated';

			switch ( $preference ) {
				case 'unsubscribe':
					$contact->unsubscribe();
                    $redirect = nonce_url_no_amp( managed_page_url( 'preferences/unsubscribe/' ), 'unsubscribe' );
                    $notice   = 'notice_unsubscribed';
					break;
				case 'confirm':
					$contact->change_marketing_preference( Preferences::CONFIRMED );
					$redirect = managed_page_url( 'preferences/confirm/' );
					break;
				case 'weekly':
					$contact->change_marketing_preference( Preferences::WEEKLY );
					break;
				case 'monthly':
					$contact->change_marketing_preference( Preferences::MONTHLY );
					break;
				case 'gdpr_delete':

					$contact->unsubscribe();
					$notice   = 'notice_unsubscribed';
					$redirect = managed_page_url( 'preferences/erase/' );
					break;
			}

			do_action( 'groundhogg/preferences/manage/preferences_updated', $contact, $preference );

			if ( $redirect ) {
				wp_redirect_with_notice( $redirect, $notice );
				die();
			}

			wp_redirect_with_notice( managed_page_url( 'preferences/profile/' ), $notice );
			die();

		}

		managed_page_head( __( 'Manage Preferences', 'groundhogg' ), 'manage' );

		$preferences = [
			'confirm'     => _x( 'I love this company, you can communicate with me whenever you feel like.', 'preferences', 'groundhogg' ),
//			'weekly'      => _x( "It's getting a bit much. Communicate with me weekly.", 'preferences', 'groundhogg' ),
//			'monthly'     => _x( 'Distance makes the heart grow fonder. Communicate with me monthly.', 'preferences', 'groundhogg' ),
			'unsubscribe' => _x( 'I no longer wish to receive any form of communication. <b>Unsubscribe me!</b>', 'preferences', 'groundhogg' )
		];

		if ( Plugin::$instance->preferences->is_gdpr_enabled() ) {
			$preferences['gdpr_delete'] = _x( 'Unsubscribe me and delete any personal information about me.', 'preferences', 'groundhogg' );
		}

		$preferences = apply_filters( 'manage_email_preferences_options', $preferences );

		$params = [
			'pk'       => $permissions_key,
			'identity' => get_url_var( 'identity' ),
		];

		do_action( 'groundhogg/preferences/manage/form/before' );

		?>
        <div class="box">
            <p>
                <b><?php printf( __( 'Managing preferences for <span class="contact-name">%s (%s)</span>.', 'groundhogg' ), $contact->get_full_name(), obfuscate_email( $contact->get_email() ) ) ?></b>
            </p>
			<?php wp_nonce_field( 'manage_email_preferences' ); ?>
			<?php do_action( 'groundhogg/preferences/manage/form/inside' ); ?>
            <div class="preference-options">
				<?php foreach ( $preferences as $preference => $text ):

					$url = add_query_arg( wp_parse_args( [
						'preference' => $preference
					], $params ), managed_page_url( 'preferences/manage/' ) )

					?>
                    <a href="<?php echo wp_nonce_url( $url, 'manage_email_preferences' ); ?>"
                       class="preference-<?php esc_attr_e( $preference ); ?>"
                    ><?php echo $text; ?></a>
				<?php endforeach; ?>
            </div>
            <p class="align-center" style="margin-top: 40px">
                <a id="gotoprofile" class="button"
                   href="<?php echo esc_url( home_url() ); ?>"><?php _e( 'Never mind, I don\'t want to change anything.', 'groundhogg' ) ?></a>
            </p>
        </div>
		<?php

		do_action( 'groundhogg/preferences/manage/form/after' );

		managed_page_footer();

		break;
	case 'unsubscribe':

		// Don't require the nonce check if one click unsub is require because it will go directly to this page.
		if ( ! wp_verify_nonce( get_request_var( '_wpnonce' ), 'unsubscribe' ) && ! is_option_enabled( 'gh_enable_one_click_unsubscribe' ) ) {
			wp_redirect( managed_page_url( 'preferences/manage/' ) );
			die();
		}

		$contact->unsubscribe();

		managed_page_head( __( 'Unsubscribed', 'groundhogg' ), 'unsubscribe' );

		?>
        <div class="box">
            <p>
                <b><?php printf( __( 'Your email address %s has just been unsubscribed.', 'groundhogg' ), obfuscate_email( $contact->get_email() ) ) ?></b>
            </p>
            <p><?php _e( 'Further interactions with our site may be interpreted as re-subscribing to our list and will result in further electronic communication.', 'groundhogg' ); ?></p>
            <a href="<?php echo wp_nonce_url( managed_page_url( 'preferences/manage/' ) . '?preference=confirm', 'manage_email_preferences' ); ?>"
               class="button preference-confirm"><?php _e( "Oops, I didn't mean to unsubscribe! <b>Re-subscribe me!</b>", 'groundhogg' ); ?></a>
        </div>
		<?php

		managed_page_footer();

		break;
	case 'confirm':

		$contact->change_marketing_preference( Preferences::CONFIRMED );
		$redirect_to = esc_url_raw( sanitize_text_field( get_url_var( 'redirect_to' ) ) );
		$redirect_to = apply_filters( 'groundhogg/confirmed/redirect_to', $redirect_to, $contact );

		if ( $redirect_to ) {
			die( wp_redirect( $redirect_to ) );
		}

		managed_page_head( __( 'Confirmed', 'groundhogg' ), 'confirm' );

		?>
        <div class="box">
            <p>
                <b><?php printf( __( 'Your email address %s has just been confirmed!', 'groundhogg' ), obfuscate_email( $contact->get_email() ) ) ?></b>
            </p>
            <p><?php printf( __( 'You will now receive electronic communication from %1$s. Should you wish to change your communication preferences you may do so at any time by clicking the <b>Manage Preferences</b> link or <b>Unsubscribe</b> link in the footer of any email sent by %1$s.', 'groundhogg' ), get_bloginfo( 'title', 'display' ) ); ?></p>
            <p>
                <a id="gotosite" class="button"
                   href="<?php echo esc_url( home_url() ); ?>"><?php printf( __( 'Return to %s', 'groundhogg' ), get_bloginfo( 'title', 'display' ) ); ?></a>
            </p>
        </div>
		<?php

		managed_page_footer();

		break;

	case 'erase':

		if ( ! wp_verify_nonce( get_request_var( '_wpnonce' ), 'erase_profile' ) ) {

			managed_page_head( __( 'Erase your profile', 'groundhogg' ), 'erase' );

			?>
            <div class="box">
                <p><b><?php _e( 'Are you sure you want to erase your profile?', 'groundhogg' ); ?></b></p>
                <p><?php _e( "Erasing your profile will mean that you will no longer receive critical updates and communications from us.", 'groundhogg' ); ?></p>
                <p><b><?php _e( "What will be erased?", 'groundhogg' ); ?></b></p>
                <ul>
                    <li><?php _e( "Your profile details.", 'groundhogg' ); ?></li>
                    <li><?php _e( "Your profile marketing history.", 'groundhogg' ); ?></li>
                    <li><?php _e( "Tracking data associated with your profile.", 'groundhogg' ); ?></li>
                </ul>
                <p><b><?php _e( "What will be <u>NOT</u> erased?", 'groundhogg' ); ?></b></p>
                <ul>
					<?php if ( $contact->get_user_id() ): ?>
                        <li><?php _e( "Your user account. <i>To erase your user account contact us.</i>", 'groundhogg' ); ?></li>
                        <li><?php _e( "Tracking data and historic details associated with your user account.", 'groundhogg' ); ?></li>
					<?php endif; ?>
                    <li><?php _e( "Associated purchase history and orders.", 'groundhogg' ); ?></li>
                    <li><?php _e( "Legal documents associated with your profile.", 'groundhogg' ); ?></li>
                </ul>
                <p>
                    <a id="gotoprofile" class="button"
                       href="<?php echo esc_url( managed_page_url( 'preferences/profile/' ) ); ?>"><?php _e( "Nevermind! Don't erase my profile.", 'groundhogg' ) ?></a>
                    <a id="eraseprofile" class="button danger right"
                       href="<?php echo esc_url( wp_nonce_url( managed_page_url( 'preferences/erase/' ), 'erase_profile' ) ); ?>"><?php _e( 'Erase Profile', 'groundhogg' ) ?></a>
                </p>
            </div>
			<?php
			managed_page_footer();

			die();
		}

		/**
		 * Before the request is made to erase the profile
		 *
		 * @param $contact Contact
		 */
		do_action( 'groundhogg/preferences/erase_profile', $contact );

		$contact->delete();

		managed_page_head( __( 'Erased', 'groundhogg' ), 'erase' );

		?>
        <div class="box">
            <p><b><?php _e( 'Your profile has been erased!', 'groundhogg' ); ?></b></p>
            <p><?php _e( 'Further interactions with our site may be interpreted as re-subscribing to our list and will result in further communication.', 'groundhogg' ); ?></p>
            <p>
                <a id="gotosite" class="button"
                   href="<?php echo esc_url( home_url() ); ?>"><?php printf( __( 'Return to %s', 'groundhogg' ), get_bloginfo( 'title', 'display' ) ); ?></a>
            </p>
        </div>
		<?php
		managed_page_footer();
		break;
endswitch;
